from typing import Generator, Optional, Iterable, Any

# ! Error Base Class
class Error(Exception):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args)
        self.args = [i for i in self.__error_text__(*args, **kwargs) if i is not None]
    
    def __error_text__(self, *args, **kwargs) -> Generator[Optional[str], Any, None]:
        yield None

# ! Language Errors
class LanguageNotExistError(Error):
    def __error_text__(self, marks: Iterable[str], *args, **kwargs):
        yield f"No language file(s) with such a mark(s) was found: {', '.join([repr(mark) for mark in marks])}"

class LanguageNotLoadedError(Error):
    def __error_text__(self, *args, **kwargs):
        yield f"The language file has not been uploaded. Use the `load` method."

# ! Exceptions
class PathNotExistsError(Exception):
    def __init__(self, path: str) -> None:
        super().__init__()
        self.args = (f"The element along the path {repr(path)} does not exist.",)

class NotBooleanError(Exception):
    def __init__(self, data: str) -> None:
        super().__init__()
        self.args = (f"The data {repr(data)} cannot be represented as a bool.",)


import os
import glob
import properties
from typing import Dict, Tuple, Optional
# > Local Imports
from .functions import formater
from .exceptions import LanguageNotExistError, LanguageNotLoadedError

# ! Child Class
class Language:
    @staticmethod
    def __load_file(filepath: str) -> Dict[str, str]:
        with open(filepath, 'r', errors='ignore', encoding='utf-8') as file:
            return properties.load(file)
    
    def __get_title(self) -> str:
        return self.__load_file(self.__name).get("language.title", f"<LTNF:{self.__mark}>")
    
    # ? Initialization
    def __init__(self, language_filepath: str) -> None:
        self.__name = os.path.abspath(language_filepath)
        self.__data, self.__loaded = None, False
        # * Checking
        if not os.path.isfile(self.__name):
            raise FileNotFoundError
        # * Metadata Loading
        self.__mark = os.path.splitext(os.path.basename(self.__name))[0].lower()
        self.__title = self.__get_title()
    
    # ? Magic Methods
    def __str__(self) -> str:
        return f"{self.__class__.__name__}({formater(name=self.__name, mark=self.__mark)})"
    
    def __repr__(self) -> str:
        return self.__str__()
    
    # ? Propertyes
    @property
    def name(self) -> str:
        return self.__name
    
    @property
    def mark(self) -> str:
        return self.__mark
    
    @property
    def title(self) -> str:
        return self.__title
    
    @property
    def loaded(self) -> bool:
        return self.__loaded
    
    # ? Main Methods
    def load(self) -> None:
        self.__data, self.__loaded = self.__load_file(self.__name), True
    
    def unload(self) -> None:
        self.__data, self.__loaded = None, False
    
    def get(self, key: str, default: Optional[str]=None) -> Optional[str]:
        if self.loaded:
            return self.__data.get(key, default)
        raise LanguageNotLoadedError()

# ! Main Class
class LanguageLoader:
    # ? Main Methods
    def __search_langs(self, dlm: str, mlm: str) -> Tuple[Language, Language]:
        dl, ml = None, None
        for lang in self.langs:
            if lang.mark == dlm:
                dl = lang
            if lang.mark == mlm:
                ml = lang
        if (dl is None) or (ml is None):
            raise LanguageNotExistError([dlm, mlm])
        dl.load()
        ml.load()
        return dl, ml
    
    # ? Initialization
    def __init__(
        self,
        langs_dirpath: str,
        main_lang_mark: str,
        default_lang_mark: str="en-eng",
    ) -> None:
        self.__name = os.path.abspath(langs_dirpath)
        self.__mlm = main_lang_mark.lower()
        self.__dlm = default_lang_mark.lower()
        # * Checking
        if not os.path.isdir(self.__name):
            raise FileNotFoundError
        # * Searching Languages
        self.langs = [Language(lp) for lp in glob.glob(os.path.join(self.__name, "??-???.properties"))]
        self.dlang, self.mlang = self.__search_langs(self.__dlm, self.__mlm)
    
    def get(self, key: str) -> str:
        return self.mlang.get(key, self.dlang.get(key, "<LTNF>"))



from textual.widgets import Button
# > Typing
from typing import Callable, Awaitable

# ! Clikable Button Class
class ClikableButton(Button):
    def __init__(self, label: str, on_click: Callable[[], Awaitable[None]], *args, **kwargs) -> None:
        super().__init__(label, *args, **kwargs)
        self.__on_click_method = on_click
    
    async def on_click(self, *args, **kwargs) -> None:
        await self.__on_click_method()


from textual.widgets import Label

# ! Main Class
class FullLabel(Label):
    def _gen(self) -> str:
        return self._chr * (self.size[0] * self.size[1])
    
    def __init__(self, char: str="-") -> None:
        super().__init__(classes="full-label")
        self._chr = char
        self.update(self._gen())
    
    async def on_resize(self) -> None:
        self.update(self._gen())


from textual.widgets import RadioSet, RadioButton
# > Typing
from typing import Optional, Callable, Generic, TypeVar

# ! Types
DT = TypeVar("DT")

# ! Data Radio Button Class
class DataRadioButton(RadioButton, Generic[DT]):
    def __init__(self, data: DT, value: bool=False, visible_data: Optional[str]=None) -> None:
        self.data = data
        self.visible_data = visible_data or str(data)
        super().__init__(visible_data, value)

# ! Data Radio Set Class
class DataRadioSet(RadioSet, Generic[DT]):
    pressed_index: int
    pressed_button: Optional[DataRadioButton[DT]]
    
    def __init__(
        self,
        on_changed: Callable[[DT], None],
        *buttons: DataRadioButton[DT],
        **kwargs
    ) -> None:
        super().__init__(*buttons, **kwargs)
        self.__on_changed_method = on_changed
    
    def _on_radio_set_changed(self, event: RadioSet.Changed) -> None:
        super()._on_radio_set_changed(event)
        self.__on_changed_method(event.pressed.data)




from pydantic import BaseModel
from textual.binding import Binding
from textual.screen import Screen
# > Typing
from typing import Optional, Generator, Type, Any
# > Local Import's
from ..codeсbase import CodecBase
from ..functions import formater

# ! Plugin Info Class
class PluginInfo(BaseModel):
    name: str
    name_id: str
    version: str
    author: str
    description: Optional[str]=None
    url: Optional[str]=None

# ! Plugin Base Class
class PluginBase:
    def __init_repr__(self) -> str:
        return f"{self.info.name} ({repr(self.info.name_id)}) [#60fdff]v{self.info.version}[/#60fdff] is [yellow]initialized[/yellow]!"
    
    def __init__(self, app, pl, info: PluginInfo) -> None:
        self.app = app
        self.pl = pl
        self.info = info
        # > Logs
        self.app.info(self.__init_repr__())
    
    def __str__(self) -> str:
        return f"{self.__class__.__name__}({formater(info=self.info)})"
    
    def __repr__(self) -> str:
        return self.__str__()
    
    # ! App Specific Functions
    def install_screen(self, name: str, screen: Screen) -> None:
        self.app.SCREENS[name] = screen
        self.app.install_screen(screen, name)
    
    def add_codecs(self, *codecs: Type[CodecBase]) -> None:
        self.app.CODECS += [ *codecs ]
    
    # ! Dev Functions
    def on_bindings(self) -> Generator[Binding, Any, None]:
        yield None
    
    def on_init(self):
        pass
    
    def on_run(self):
        pass
    
    async def on_compose(self):
        pass
    
    async def on_quit(self):
        pass



from textual.widgets import Label
from textual.containers import Center, Vertical, Horizontal
from rich.progress import Progress, BarColumn, TextColumn
# > Typing
from typing import Callable
# > Local Imports
from .Buttons import ClikableButton

# ! RheostatBar Class
class RheostatBar(Label):
    def __match_bar_width(self) -> int:
        bar_size = self.size[0] - (len(f"{self.__min_value}{self.__max_value}")+(len(self.__mark)*2)+2)
        if bar_size <= 0:
            bar_size = 4
        return bar_size
    
    def __init__(
        self,
        value: int=0,
        min_value: int=0,
        max_value: int=100,
        mark: str=""
    ) -> None:
        self.__value, self.__min_value, self.__max_value = value, min_value, max_value
        self.__mark = mark
        self.__process = Progress(
            TextColumn(f"{self.__min_value}{self.__mark} "),
            BarColumn(4),
            TextColumn(" {task.total}"+self.__mark)
        )
        self.__tid = self.__process.add_task("", completed=self.__value, total=self.__max_value)
        super().__init__(self.__process, classes="rheostat-bar")
    
    # * Property
    @property
    def value(self) -> int:
        """Currect value."""
        return self.__value
    @value.setter
    def value(self, value: int) -> None:
        self.__process.update(self.__tid, completed=value)
        self.__value = value
        self.update(self.__process)
    
    @property
    def min_value(self) -> int:
        """Minimal value."""
        return self.__min_value
    @min_value.setter
    def min_value(self, value: int) -> None:
        self.__process.columns[0].text_format = f"{value}{self.__mark} "
        self.__min_value = value
        self.update(self.__process)
    
    @property
    def max_value(self) -> int:
        """Maximal value."""
        return self.__max_value
    @max_value.setter
    def max_value(self, value: int) -> None:
        self.__process.columns[2].text_format = " {task.total}"+self.__mark
        self.__max_value = value
        self.update(self.__process)
    
    @property
    def mark(self) -> str:
        """Marking of value.
        Formating:
        ```
        f"{value}{mark}"
        ```
        """
        return self.__mark
    @mark.setter
    def mark(self, value: str) -> None:
        self.__mark = value
        self.max_value = self.max_value
        self.min_value = self.min_value
        self.update(self.__process)
    
    # * Methods
    async def on_resize(self) -> None:
        bar_width = self.__match_bar_width()
        if self.__process.columns[1].bar_width != bar_width:
            self.__process.columns[1].bar_width = bar_width
        self.update(self.__process)

class Rheostat(Vertical):
    async def __click_plus(self) -> None:
        if (self.bar.value + self.__advance_value) <= self.bar.max_value:
            self.bar.value = self.bar.value + self.__advance_value
            self.label.update(f" {self.bar.value}{self.bar.mark} ")
            self.__on_change_value_method(self.bar.value)
    
    async def __click_minus(self) -> None:
        if (self.bar.value - self.__advance_value) >= self.bar.min_value:
            self.bar.value = self.bar.value - self.__advance_value
            self.label.update(f" {self.bar.value}{self.bar.mark} ")
            self.__on_change_value_method(self.bar.value)
    
    def __init__(
        self,
        on_change_value: Callable[[int], None],
        value: int=0,
        advance_value: int=1,
        min_value: int=0,
        max_value: int=100,
        mark: str=""
    ) -> None:
        self.bar = RheostatBar(value, min_value, max_value, mark)
        self.label = Label(f" {self.bar.value}{self.bar.mark} ")
        self.__advance_value = advance_value
        self.__on_change_value_method = on_change_value
        super().__init__(
            Center(self.bar),
            Horizontal(
                ClikableButton("-", self.__click_minus),
                self.label,
                ClikableButton("+", self.__click_plus),
                classes="rheostat-horizontal"
            ),
            classes="rheostat"
        )


import os
from rich.console import Console
# > Typing
from typing import List
# > Local Import's
from .functions import *
from ..pluginbase import PluginInfo
from ..pluginloader import PluginLoaderConfigManager, PluginLoader
from ...units import PLUGINS_CONFIG_PATH, PLUGINS_DIRPATH

# ! Functions
def init_config() -> None:
    os.makedirs(PLUGINS_DIRPATH, mode=0o755, exist_ok=True)
    os.makedirs(os.path.dirname(PLUGINS_CONFIG_PATH), mode=0o755, exist_ok=True)
    plugin_config, plugins_infos = PluginLoaderConfigManager(PLUGINS_CONFIG_PATH), get_plugins_info()
    add_plugins(plugin_config, *plugins_infos)

def is_config_inited() -> bool:
    try:
        assert os.path.exists(PLUGINS_DIRPATH)
        assert os.path.exists(os.path.dirname(PLUGINS_CONFIG_PATH))
        assert os.path.exists(PLUGINS_CONFIG_PATH)
    except:
        return False
    return True

def add_plugins(
    plugin_config: PluginLoaderConfigManager,
    *infos: PluginInfo
) -> None:
    for info in infos:
        if not plugin_config.exists_plugin(info):
            plugin_config.add_plugin(info)

def get_plugins_info() -> List[PluginInfo]:
    plugins_infos = []
    for plugin_init_path, plugin_info_path, plugin_deps_path in PluginLoader.search_plugins_paths():
        try:
            plugins_infos.append(PluginLoader.load_plugin_info(plugin_info_path))
        except:
            pass
    return plugins_infos

def is_plugin_dirpath(dirpath: str) -> bool:
    try:
        assert os.path.exists(dirpath)
        assert os.path.exists(os.path.join(dirpath, "info.json"))
        assert os.path.exists(os.path.join(dirpath, "__init__.py"))
        PluginInfo.model_validate_json(open(os.path.join(dirpath, "info.json"), "rb").read())
    except:
        return False
    return True

# ! Other
def raise_exception(console: Console, exc_type: Exception, *args, **kwargs) -> None:
    try: raise exc_type(*args, **kwargs)
    except: console.print_exception()


import os
import asyncio
import aiofiles
import subprocess
from tempfile import mkstemp
# > IO's
from urlopen2 import URLFile
# > Sound Works
from playsoundsimple import Sound
from playsoundsimple.sound import FPType, getfp
from playsoundsimple.units import DEFAULT_SOUND_FONTS_PATH
from playsoundsimple.exceptions import FileTypeError
# > Typing
from typing import Optional

# ! Vars
FLUID_SYNTH_PATH = 'fluidsynth'

# ! Main Class
class AnySound(Sound):
    @staticmethod
    async def aio_from_midi(
        fp: FPType,
        sound_fonts_path: Optional[str]=None,
        **kwargs
    ):
        path, is_temp = getfp(fp, filetype=".midi")
        sound_fonts_path = sound_fonts_path or DEFAULT_SOUND_FONTS_PATH
        if path is None:
            raise FileTypeError(fp)
        npath = mkstemp(suffix=".wav")[1]
        
        process = await asyncio.create_subprocess_exec(
            FLUID_SYNTH_PATH, "-ni", sound_fonts_path, path, "-F", npath, "-q",
            stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
        )
        await process.wait()
        
        if is_temp:
            try:
                os.remove(path)
            except:
                pass
        
        return AnySound(npath, **{"is_temp": True, **kwargs})
    
    @staticmethod
    def from_midi(
        fp: FPType,
        sound_fonts_path: Optional[str]=None,
        **kwargs
    ):
        path, is_temp = getfp(fp, filetype=".midi")
        sound_fonts_path = sound_fonts_path or DEFAULT_SOUND_FONTS_PATH
        if path is None:
            raise FileTypeError(fp)
        npath = mkstemp(suffix=".wav")[1]
        
        subprocess.call(
            [FLUID_SYNTH_PATH, "-ni", sound_fonts_path, path, "-F", npath, "-q"],
            stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
        )
        
        if is_temp:
            try:
                os.remove(path)
            except:
                pass
        
        return AnySound(npath, **{"is_temp": True, **kwargs})
    
    @staticmethod
    def from_url(
        url: str,
        download_buffer_size: int=65536,
        **kwargs
    ):
        path = mkstemp(suffix=".bin")[1]
        
        with URLFile(url) as urlfile:
            with open(path, "wb") as tempfile:
                while len(data:=urlfile.read(download_buffer_size)) != 0:
                    tempfile.write(data)
        
        return AnySound(path, **{"is_temp": True, **kwargs})
    
    @staticmethod
    async def aio_from_url(
        url: str,
        download_buffer_size: int=65536,
        **kwargs
    ):
        path = mkstemp(suffix=".bin")[1]
        
        with URLFile(url) as urlfile:
            async with aiofiles.open(path, "wb") as tempfile:
                while len(data:=urlfile.read(download_buffer_size)) != 0:
                    await tempfile.write(data)
        
        return AnySound(path, **{"is_temp": True, **kwargs})



import os
import sys
import glob
import asyncio
from pathlib import Path
from pydantic import BaseModel
from rich.console import Console
from textual.binding import Binding
# > ImportLib
from importlib.util import spec_from_file_location, module_from_spec
# > Typing
from types import ModuleType
from typing import Optional, Dict, Union, Any, List, Tuple, Type, Generator
# > Local Import's
from .pipw import pip
from .pluginbase import PluginInfo, PluginBase
from ..functions import aiter
from ..units import (
    PLUGINS_DIRPATH,
    PLUGINS_CONFIG_PATH,
    GLOB_PLUGINS_INFO_SEARCH,
    GLOB_PLUGINS_INIT_SEARCH,
    GLOB_PLUGINS_DEPS_SEARCH
)

# ! Vars
console = Console()

# ! Types
class PluginModuleType(ModuleType):
    plugin_main: Type[PluginBase]

# ! Functions
def get_module_info(path: str):
    if os.path.basename(path) == "__init__.py":
        return os.path.basename(os.path.dirname(path)), path
    return os.path.basename(path), path

def get_submodules_locations(init_path: str) -> List[str]:
    return [ os.path.dirname(init_path) ]

def load_module(init_path: str) -> PluginModuleType:
    module_name, module_location = get_module_info(init_path)
    module_spec = spec_from_file_location(
        module_name,
        module_location,
        submodule_search_locations=get_submodules_locations(init_path)
    )
    module = module_from_spec(module_spec)
    
    sys.modules[module_spec.name] = module
    # TODO: Temporary option as there is a risk of replacing existing modules
    # TODO: Make an environment-module to surround all these modules for security
    
    module_spec.loader.exec_module(module)
    return module

def plugin_from_module(app, pl, info: PluginInfo, module: PluginModuleType) -> PluginBase:
    return module.plugin_main(app, pl, info)

# ! Plugin Loader Config
class PluginLoaderConfigModel(BaseModel):
    plugins_enable: Dict[str, bool] = {}

class PluginLoaderConfigManager:
    @staticmethod
    def dump(path: str, data: PluginLoaderConfigModel) -> None:
        with open(path, 'w') as file:
            file.write(data.model_dump_json())
    
    @staticmethod
    def load(path: str, default_data: Dict[str, Any]) -> PluginLoaderConfigModel:
        try:
            with open(path, 'rb') as file:
                data = file.read()
            return PluginLoaderConfigModel.model_validate_json(data)
        except:
            return default_data
    
    def refresh(self) -> None:
        self.dump(self.filepath, self.config)
    
    def __init__(self, path: str) -> None:
        self.filepath = Path(path)
        
        default_data = PluginLoaderConfigModel().model_dump()
        if self.filepath.exists():
            self.config = self.load(self.filepath, default_data)
            config_temp = default_data.copy()
            config_temp.update(self.config)
            self.config = PluginLoaderConfigModel.model_validate(config_temp)
        else:
            self.config = PluginLoaderConfigModel.model_validate(default_data)
        self.refresh()
    
    def exists_plugin(self, info: PluginInfo) -> bool:
        return info.name_id in self.config.plugins_enable.keys()
    
    def exists_plugin_by_name_id(self, name_id: str) -> bool:
        return name_id in self.config.plugins_enable.keys()
    
    def add_plugin(self, info: PluginInfo) -> None:
        self.config.plugins_enable[info.name_id] = True
        self.refresh()
    
    def is_enable_plugin(self, info: PluginInfo) -> bool:
        for name_id, enable in self.config.plugins_enable.items():
            if name_id == info.name_id:
                return enable
        return False
    
    def disable_plugin(self, info: PluginInfo) -> None:
        self.config.plugins_enable[info.name_id] = False
        self.refresh()
    
    def disable_plugin_by_name_id(self, name_id: str) -> None:
        self.config.plugins_enable[name_id] = False
        self.refresh()
    
    def enable_plugin(self, info: PluginInfo) -> None:
        self.config.plugins_enable[info.name_id] = True
        self.refresh()
    
    def enable_plugin_by_name_id(self, name_id: str) -> None:
        self.config.plugins_enable[name_id] = True
        self.refresh()

# ! Plugin Loader Class
class PluginLoader:
    __title__: str = "PluginLoader"
    __version__: str = "0.3.0"
    __author__: str = "Romanin"
    __email__: str = "semina054@gmail.com"
    
    def __init__(
        self,
        app,
        plugins_dirpath: Optional[Union[str, Path]]=None,
        plugins_config_path: Optional[Union[str, Path]]=None,
        *args,
        **kwargs
    ) -> None:
        self.app = app
        self.plugins_dirpath = Path(os.path.abspath(plugins_dirpath or PLUGINS_DIRPATH))
        self.plugins_config_path = Path(os.path.abspath(plugins_config_path or PLUGINS_CONFIG_PATH))
        
        # * Create plugins directory
        os.makedirs(self.plugins_dirpath, 0o755, True)
        
        # * Config Initializing
        self.config = PluginLoaderConfigManager(self.plugins_config_path)
        
        # * Vars
        self.on_plugins: List[PluginBase] = []
        self.off_plugins: List[PluginInfo] = []
        self.error_plugins: List[Tuple[str, str]] = []
        
        # * Logging
        self.app.info("---")
    
    @staticmethod
    async def aio_search_plugins_paths():
        init_search, info_search, deps_search =  \
            glob.glob(GLOB_PLUGINS_INIT_SEARCH), \
            glob.glob(GLOB_PLUGINS_INFO_SEARCH), \
            glob.glob(GLOB_PLUGINS_DEPS_SEARCH)
        async for init_path in aiter(init_search):
            _init_dirpath = os.path.dirname(init_path)
            info_path, deps_path = None, None
            async for _info_path in aiter(info_search):
                _info_dirpath = os.path.dirname(_info_path)
                if _info_dirpath == _init_dirpath:
                    info_path = _info_path
                    break
            async for _deps_path in aiter(deps_search):
                _deps_dirpath = os.path.dirname(_deps_path)
                if _deps_dirpath == _init_dirpath:
                    deps_path = _deps_path
                    break
            if info_path is not None:
                yield init_path, info_path, deps_path
                await asyncio.sleep(0)
    
    @staticmethod
    def search_plugins_paths():
        init_search, info_search, deps_search =  \
            glob.glob(GLOB_PLUGINS_INIT_SEARCH), \
            glob.glob(GLOB_PLUGINS_INFO_SEARCH), \
            glob.glob(GLOB_PLUGINS_DEPS_SEARCH)
        for init_path in init_search:
            _init_dirpath = os.path.dirname(init_path)
            info_path, deps_path = None, None
            for _info_path in info_search:
                _info_dirpath = os.path.dirname(_info_path)
                if _info_dirpath == _init_dirpath:
                    info_path = _info_path
                    break
            for _deps_path in deps_search:
                _deps_dirpath = os.path.dirname(_deps_path)
                if _deps_dirpath == _init_dirpath:
                    deps_path = _deps_path
                    break
            if info_path is not None:
                yield init_path, info_path, deps_path
    
    @staticmethod
    def load_plugin_info(path: str) -> PluginInfo:
        with open(path, 'rb') as file:
            data = file.read()
        return PluginInfo.model_validate_json(data)
    
    # ! On Init Method
    def on_init(self) -> None:
        self.app.info(f"{self.__title__} [#60fdff]v{self.__version__}[/#60fdff] from {self.__author__} ({self.__email__})", in_console=True)
        plugins_paths = list(self.search_plugins_paths())
        self.app.info(f"Found plugins        : {repr([os.path.basename(os.path.dirname(i[0])) for i in plugins_paths])}", in_console=True)
        self.app.info(f"Initialization plugins...", in_console=True)
        for init_path, info_path, deps_path in plugins_paths:
            info = None
            try:
                info = self.load_plugin_info(info_path)
                if not self.config.exists_plugin(info):
                    self.config.add_plugin(info)
                    self.app.info(f"{info.name} ({repr(info.name_id)}) > New plugin added to config!", in_console=True)
                if self.config.is_enable_plugin(info):
                    self.app.info(f"{info.name} ({repr(info.name_id)}) > Plugin is [green]enabled[/green]!", in_console=True)
                    if deps_path is not None:
                        self.app.info(f"{info.name} ({repr(info.name_id)}) > Installing plugin dependencies...", in_console=True)
                        pip.install_requirements(deps_path, True)
                        self.app.info(f"{info.name} ({repr(info.name_id)}) > Installed!", in_console=True)
                    self.app.info(f"{info.name} ({repr(info.name_id)}) > Importing in SeaPlayer...", in_console=True)
                    plugin_module = load_module(init_path)
                    plugin = plugin_from_module(self.app, self, info, plugin_module)
                    self.app.info(f"{info.name} ({repr(info.name_id)}) > Imported!", in_console=True)
                    try:
                        plugin.on_init()
                    except:
                        self.app.error(f"Failed to do [green]`on_init`[/green] in: {plugin.info}", in_console=True)
                    self.on_plugins.append(plugin)
                else:
                    self.app.info(f"{info.name} ({repr(info.name_id)}) > Plugin is [red]disabled[/red]!", in_console=True)
                    self.off_plugins.append(info)
            except Exception as e:
                self.error_plugins.append( (info_path, init_path) )
                if info is not None:
                    self.app.error(f"Failed to load plugin: {repr(info)}", in_console=True)
                else:
                    self.app.error(f"Failed to load plugin: {repr(os.path.basename(os.path.dirname(info_path)))}", in_console=True)
                raise e
        self.app.info(f"Plugins loaded ([green]ON [/green]) : {repr(self.on_plugins)}", in_console=True)
        self.app.info(f"Plugins loaded ([red]OFF[/red]) : {repr(self.off_plugins)}", in_console=True)
        self.app.info(f"---", in_console=True)
    
    # ! App Specific Methods
    def on_bindings(self) -> Generator[Binding, Any, None]:
        yield None
        for plugin in self.on_plugins:
            for binding in plugin.on_bindings():
                if binding is not None:
                    yield binding
    
    # ! On Methods
    def on_run(self) -> None:
        for i in self.on_plugins:
            try:
                i.on_run()
            except:
                self.app.error(f"Failed to do [green]`on_run`[/green] in: {i}")
    
    async def on_compose(self) -> None:
        async for i in aiter(self.on_plugins):
            try:
                await i.on_compose()
            except:
                self.app.error(f"Failed to do [green]`await on_compose`[/green] in: {i}")
    
    async def on_quit(self) -> None:
        async for i in aiter(self.on_plugins):
            try:
                await i.on_quit()
            except:
                self.app.error(f"Failed to do [green]`await on_quit`[/green] in: {i}")



import os
import asyncio
import aiofiles
from io import BytesIO
# > Image Works
from PIL import Image
# > Typing
from typing_extensions import deprecated
from typing import Literal, Tuple, Optional, Iterable, TypeVar, AsyncGenerator, Any
# > Local Imports
from .codeсbase import CodecBase

# ! Types
T = TypeVar("T")

# ! Async Functions
async def aiter(it: Iterable[T]) -> AsyncGenerator[T, Any]:
    for i in it:
        await asyncio.sleep(0)
        yield i

async def get_bar_status() -> Tuple[str, Optional[float], Optional[float]]:
    return "", None, None

async def aio_is_midi_file(filepath: str):
    async with aiofiles.open(filepath, 'rb') as file:
        return await file.read(4) == b"MThd"

async def aio_check_status_code(sound: CodecBase) -> Literal[0, 1, 2]:
    if sound.playing:
        return 2 if sound.paused else 1
    else:
        return 0

# ! Exceptions Rich
def rich_exception(exc: Exception) -> str:
    return f"[red]{exc.__class__.__name__}[/red]: {exc.__str__()}"

# ! Functions
def formater(**kwargs) -> str:
    return ", ".join([f"{key}={repr(value)}" for key, value in kwargs.items()])

@deprecated("Use `seaplayer.seaplayer.SeaPlayer.get_sound_tstatus`.")
def check_status(sound: CodecBase) -> Literal["Stoped", "Playing", "Paused"]:
    if sound.playing:
        if sound.paused:
            return "Paused"
        else:
            return "Playing"
    return "Stoped"

def get_sound_basename(sound: CodecBase) -> str:
    if sound.title is not None:
        if sound.artist is not None:
            return f"{sound.artist} - {sound.title}"
        return f"{sound.title}"
    if sound.name is not None:
        try:
            return f"{os.path.basename(sound.name)}"
        except:
            return sound.name
    else:
        return "<memory>"

def image_from_bytes(data: Optional[bytes]) -> Optional[Image.Image]:
    if data is not None:
        return Image.open(BytesIO(data))



from textual.app import ComposeResult
from textual.screen import Screen
from textual.widgets import Header, Footer
# > Typing
try:
    from sounddevice import query_devices, query_hostapis
    INIT_SOUNDDEVICE = True
except:
    INIT_SOUNDDEVICE = False
from typing import Optional, Literal, Union, Tuple, Dict, List, Any
# > Local Imports
from ..languages import LanguageLoader
from ..types import Converter
from ..modules.colorizer import richefication
from ..objects import (
    Nofy,
    InputField,
    DataOptionList, DataOption,
    DataRadioSet, DataRadioButton,
    ConfigurateList, ConfigurateListItem
)

# ! Vars
conv = Converter()

# ! Functions
if INIT_SOUNDDEVICE:
    def generate_devices_options(currect: Optional[int]=None):
        hosts: List[Dict[str, Any]] = [_1 for _1 in query_hostapis()]
        devices: List[Dict[str, Any]] = [_2 for _2 in query_devices()]
        devices_options: List[DataOption] = []
        devices_options.append(DataOption("([cyan]*[/cyan]) [yellow]Auto[/yellow]", device_index=None))
        
        for device in devices:
            if device["max_output_channels"] > 0:
                try:
                    format_data = dict(
                        device_name=device["name"], 
                        device_index=device["index"],
                        hostapi_index=device["hostapi"],
                        hostapi_name=hosts[device["hostapi"]]["name"]
                    )
                    devices_options.append(
                        DataOption(
                            "([cyan]{device_index}[/cyan]) [yellow]{device_name}[/yellow] \[[green]{hostapi_name}[/green]]".format(**format_data),
                            device_index=device["index"]
                        )
                    )
                except:
                    pass
        for d in devices_options:
            if d.data["device_index"] == currect:
                d.selected = True
        return devices_options

# ! Main Class
class Configurate(Screen):
    """This Configurate Menu screen."""
    # ! Propertyes
    @property
    def ll(self) -> LanguageLoader:
        return self.app.ll
    
    # ! Textual Settings
    BINDINGS = [("escape", "app.pop_screen", "Back")]
    
    # ! Nofy Functions
    def nofy(
        self,
        text: str,
        life_time: float=3,
        dosk: Literal["bottom", "left", "right", "top"]="top"
    ) -> None:
        self.screen.mount(Nofy(text, life_time, dosk))
    
    async def aio_nofy(
        self,
        text: str,
        life_time: float=3,
        dosk: Literal["bottom", "left", "right", "top"]="top"
    ) -> None:
        await self.screen.mount(Nofy(text, life_time, dosk))
    
    # ! Update Placeholder from InputField
    def _upfif(self, attr_name: str) -> str:
        return self.ll.get("words.currect")+": " + str(eval(f"self.{attr_name}"))
    
    def gupfif(self, attr_name: str):
        return lambda: self._upfif(attr_name)
    
    # ! Update App Config
    async def _auac(self, attr_name: str, input: InputField, value: str) -> None:
        exec(f"self.{attr_name} = value")
        self.app.update_bindings()
        await self.aio_nofy("Saved!")
    
    def _caa(self, attr_name: str, value: str) -> None:
        exec(f"self.{attr_name} = value")
        self.app.update_bindings()
        self.nofy("Saved!")
    
    def _gaa(self, attr_name: str) -> Any:
        return eval(f"self.{attr_name}")
    
    if INIT_SOUNDDEVICE:
        def gucsdi(self):
            async def n_ucsdi(option: DataOption) -> None:
                self.app.config.output_sound_device_id = option.data.get("device_index", None)
                await self.aio_nofy("Saved!")
            return n_ucsdi
    
    def guac(self, attr_name: str):
        async def an_uac(input: InputField, value: str) -> None:
            await self._auac(attr_name, input, value)
        return an_uac
    
    # ! Configurator Generators
    def create_configurator_type(
        self,
        attr_name: str,
        group: str="Option",
        title: str="",
        desc: str="",
        _type: type=str,
        type_alias: type=str,
        restart_required: bool=True
    ) -> ConfigurateListItem:
        return ConfigurateListItem(
            InputField(
                conv=conv.gen_aio_conv(_type),
                submit=self.guac(attr_name),
                update_placeholder=self.gupfif(attr_name)
            ),
            title="[red]{"+group+"}[/red]: "+f"{title} ({richefication(type_alias)})",
            desc=desc+(f" [red]({self.ll.get('words.restart_required')})[/red]" if restart_required else ""),
            height=5
        )
    
    def create_configurator_keys(
        self,
        attr_name: str,
        title: str="",
        desc: str="",
        restart_required: bool=True
    ) -> ConfigurateListItem:
        return ConfigurateListItem(
            InputField(
                submit=self.guac(attr_name),
                update_placeholder=self.gupfif(attr_name)
            ),
            title="[red]{Keys}[/red]: "+f"{title} ({richefication(str)})",
            desc=desc+(f" [red]({self.ll.get('words.restart_required')})[/red]" if restart_required else ""),
            height=5
        )
    
    def create_configurator_literal(
        self,
        attr_name: str,
        values: List[Union[Any, Tuple[Any, str]]],
        group: str="Option",
        title: str="",
        desc: str="",
        restart_required: bool=True
    ) -> ConfigurateListItem:
        on_changed_method = lambda v: self._caa(attr_name, v)
        default_value = [v[0] if isinstance(v, tuple) else v for v in values].index(self._gaa(attr_name))
        buttons = []
        for idx, d in enumerate(values):
            drb = \
                DataRadioButton(d[0], idx==default_value, d[1]) \
            if isinstance(d, tuple) else \
                DataRadioButton(d, idx==default_value)
            buttons.append(drb)
        return ConfigurateListItem(
            DataRadioSet(on_changed_method, *buttons),
            title="[red]{"+group+"}[/red]: "+title,
            desc=desc + f" [red]({self.ll.get('words.restart_required')})[/red]" if restart_required else "",
            height=len(values)+4
        )
    
    if INIT_SOUNDDEVICE:
        def create_configurator_sound_devices(self):
            dos = generate_devices_options(self.app.config.output_sound_device_id)
            options_list = DataOptionList(
                *dos,
                group="SoundDevicesSelect",
                after_selected=self.gucsdi()
            )
            return ConfigurateListItem(
                options_list,
                title="[red]{"+self.ll.get("configurate.sound")+"}[/]: "+self.ll.get("configurate.sound.output_device"),
                desc=self.ll.get("configurate.sound.output_device.desc"),
                height=len(dos)+4
            )
    
    # ! Configurate Main Functions
    def compose(self) -> ComposeResult:
        self.congigurate_list = ConfigurateList()
        self.congigurate_list.border_title = self.ll.get("configurate")
        yield Header()
        with self.congigurate_list:
            yield self.create_configurator_literal(
                "app.config.lang",
                [(lang.mark, lang.title) for lang in self.ll.langs],
                self.ll.get("configurate.main"),
                self.ll.get("configurate.main.lang"),
                self.ll.get("configurate.main.lang.desc")
            )
            yield self.create_configurator_type(
                "app.config.sound_font_path",
                self.ll.get("configurate.sound"),
                self.ll.get("configurate.sound.font_path"),
                self.ll.get("configurate.sound.font_path.desc"),
                conv.optional(conv.filepath), Optional[str], False
            )
            if INIT_SOUNDDEVICE:
                yield self.create_configurator_sound_devices()
            yield self.create_configurator_literal(
                "app.config.image_update_method",
                [
                    ("sync", self.ll.get("configurate.image.update_method.sync")),
                    ("async", self.ll.get("configurate.image.update_method.async"))
                ],
                self.ll.get("configurate.image"),
                self.ll.get("configurate.image.update_method"),
                self.ll.get("configurate.image.update_method.desc")
            )
            yield self.create_configurator_literal(
                "app.config.image_resample_method",
                [
                    ("nearest", self.ll.get("configurate.image.resample_method.nearest")),
                    ("bilinear", self.ll.get("configurate.image.resample_method.bilinear")),
                    ("bicubic", self.ll.get("configurate.image.resample_method.bicubic")),
                    ("lanczos", self.ll.get("configurate.image.resample_method.lanczos")),
                    ("hamming", self.ll.get("configurate.image.resample_method.hamming")),
                    ("box", self.ll.get("configurate.image.resample_method.box")),
                ],
                self.ll.get("configurate.image"),
                self.ll.get("configurate.image.resample_method"),
                self.ll.get("configurate.image.resample_method.desc")
            )
            yield self.create_configurator_type(
                "app.config.volume_change_percent",
                self.ll.get("configurate.playback"),
                self.ll.get("configurate.playback.volume_change_percent"),
                self.ll.get("configurate.playback.volume_change_percent.desc"),
                float, float, False
            )
            yield self.create_configurator_type(
                "app.config.rewind_count_seconds",
                self.ll.get("configurate.playback"),
                self.ll.get("configurate.playback.rewind_count_seconds"),
                self.ll.get("configurate.playback.rewind_count_seconds.desc"),
                int, int, False
            )
            yield self.create_configurator_type(
                "app.config.max_volume_percent",
                self.ll.get("configurate.playback"),
                self.ll.get("configurate.playback.max_volume_percent"),
                self.ll.get("configurate.playback.max_volume_percent.desc"),
                float, float, False
            )
            yield self.create_configurator_literal(
                "app.config.recursive_search",
                [
                    (True, self.ll.get("words.on")),
                    (False, self.ll.get("words.off"))
                ],
                self.ll.get("configurate.playlist"),
                self.ll.get("configurate.playlist.recursive_search"),
                self.ll.get("configurate.playlist.recursive_search.desc"),
                False
            )
            yield self.create_configurator_literal(
                "app.config.log_menu_enable",
                [
                    (True, self.ll.get("words.on")),
                    (False, self.ll.get("words.off"))
                ],
                self.ll.get("configurate.debug"),
                self.ll.get("configurate.debug.log_menu_enable"),
                self.ll.get("configurate.debug.log_menu_enable.desc"),
                False
            )
            # ! Keys
            yield self.create_configurator_keys(
                "app.config.key_quit",
                self.ll.get("configurate.keys.quit"),
                self.ll.get("configurate.keys.quit.desc"),
                False
            )
            yield self.create_configurator_keys(
                "app.config.key_rewind_forward",
                self.ll.get("configurate.keys.rewind_forward"),
                self.ll.get("configurate.keys.rewind_forward.desc"),
                False
            )
            yield self.create_configurator_keys(
                "app.config.key_rewind_back",
                self.ll.get("configurate.keys.rewind_back"),
                self.ll.get("configurate.keys.rewind_back.desc"),
                False
            )
            yield self.create_configurator_keys(
                "app.config.key_volume_up",
                self.ll.get("configurate.keys.volume_plus"),
                self.ll.get("configurate.keys.volume_plus.desc"),
                False
            )
            yield self.create_configurator_keys(
                "app.config.key_volume_down",
                self.ll.get("configurate.keys.volume_minus"),
                self.ll.get("configurate.keys.volume_minus.desc"),
                False
            )
        yield Footer()
